# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import io
import pathlib
import warnings
from collections.abc import Callable, Generator
from functools import partial, wraps
from typing import Any, TypeVar

import h5py
import numpy as np
import sciline as sl
import scipp as sc

from .types import (
    DetectorIndex,
    DetectorName,
    FilePath,
    NMXDetectorMetadata,
    NMXExperimentMetadata,
    NMXReducedDataGroup,
)


def _fallback_compute_positions(dg: sc.DataGroup) -> sc.DataGroup:
    import warnings

    import scippnexus as snx

    warnings.warn(
        "Using fallback compute_positions due to empty log entries. "
        "This may lead to incorrect results. Please check the data carefully."
        "The fallback will replace empty logs with a scalar value of zero.",
        UserWarning,
        stacklevel=2,
    )

    empty_transformations = [
        transformation
        for transformation in dg['depends_on'].transformations.values()
        if 'time' in transformation.value.dims
        and transformation.sizes['time'] == 0  # empty log
    ]
    for transformation in empty_transformations:
        orig_value = transformation.value
        orig_value = sc.scalar(0, unit=orig_value.unit, dtype=orig_value.dtype)
        transformation.value = orig_value
    return snx.compute_positions(dg, store_transform='transform_matrix')


def _compute_positions(
    dg: sc.DataGroup, auto_fix_transformations: bool = False
) -> sc.DataGroup:
    """Compute positions of the data group from transformations.

    Wraps the `scippnexus.compute_positions` function
    and provides a fallback for cases where the transformations
    contain empty logs.

    Parameters
    ----------
    dg:
        Data group containing the transformations and data.
    auto_fix_transformations:
        If `True`, it will attempt to fix empty transformations.
        It will replace them with a scalar value of zero.
        It is because adding a time dimension will make it not possible
        to compute positions of children due to time-dependent transformations.

    Returns
    -------
    :
        Data group with computed positions.

    Warnings
    --------
    If `auto_fix_transformations` is `True`, it will warn about the fallback
    being used due to empty logs or scalar transformations.
    This is because the fallback may lead to incorrect results.

    """
    import scippnexus as snx

    try:
        return snx.compute_positions(dg, store_transform='transform_matrix')
    except ValueError as e:
        if auto_fix_transformations:
            return _fallback_compute_positions(dg)
        raise e


def _create_dataset_from_string(*, root_entry: h5py.Group, name: str, var: str) -> None:
    root_entry.create_dataset(name, dtype=h5py.string_dtype(), data=var)


def _create_dataset_from_var(
    *,
    root_entry: h5py.Group,
    var: sc.Variable,
    name: str,
    long_name: str | None = None,
    compression: str | None = None,
    compression_opts: int | tuple[int, int] | None = None,
    chunks: tuple[int, ...] | int | bool | None = None,
    dtype: Any = None,
) -> h5py.Dataset:
    compression_options = {}
    if compression is not None:
        compression_options["compression"] = compression
    if compression_opts is not None:
        compression_options["compression_opts"] = compression_opts

    dataset = root_entry.create_dataset(
        name,
        data=var.values if dtype is None else var.values.astype(dtype, copy=False),
        chunks=chunks,
        **compression_options,
    )
    if var.unit is not None:
        dataset.attrs["units"] = str(var.unit)
    if long_name is not None:
        dataset.attrs["long_name"] = long_name
    return dataset


@wraps(_create_dataset_from_var)
def _create_compressed_dataset(*args, **kwargs):
    """Create dataset with compression options.

    It will try to use ``bitshuffle`` for compression if available.
    Otherwise, it will fall back to ``gzip`` compression.

    [``Bitshuffle/LZ4``](https://github.com/kiyo-masui/bitshuffle)
    is used for convenience.
    Since ``Dectris`` uses it for their Nexus file compression,
    it is compatible with DIALS.
    ``Bitshuffle/LZ4`` tends to give similar results to
    GZIP and other compression algorithms with better performance.
    A naive implementation of bitshuffle/LZ4 compression,
    shown in [issue #124](https://github.com/scipp/essnmx/issues/124),
    led to 80% file reduction (365 MB vs 1.8 GB).

    """
    try:
        import bitshuffle.h5

        compression_filter = bitshuffle.h5.H5FILTER
        default_compression_opts = (0, bitshuffle.h5.H5_COMPRESS_LZ4)
    except ImportError:
        warnings.warn(
            UserWarning(
                "Could not find the bitshuffle.h5 module from bitshuffle package. "
                "The bitshuffle package is not installed or only partially installed. "
                "Exporting to NeXus files with bitshuffle compression is not possible."
            ),
            stacklevel=2,
        )
        compression_filter = "gzip"
        default_compression_opts = 4

    return _create_dataset_from_var(
        *args,
        **kwargs,
        compression=compression_filter,
        compression_opts=default_compression_opts,
    )


def _create_root_data_entry(file_obj: h5py.File) -> h5py.Group:
    nx_entry = file_obj.create_group("NMX_data")
    nx_entry.attrs["NX_class"] = "NXentry"
    nx_entry.attrs["default"] = "data"
    nx_entry.attrs["name"] = "NMX"
    nx_entry["name"] = "NMX"
    nx_entry["definition"] = "TOFRAW"
    return nx_entry


def _create_sample_group(data: sc.DataGroup, nx_entry: h5py.Group) -> h5py.Group:
    nx_sample = nx_entry.create_group("NXsample")
    nx_sample["name"] = data['sample_name'].value
    _create_dataset_from_var(
        root_entry=nx_sample,
        var=data['crystal_rotation'],
        name='crystal_rotation',
        long_name='crystal rotation in Phi (XYZ)',
    )
    return nx_sample


def _create_instrument_group(data: sc.DataGroup, nx_entry: h5py.Group) -> h5py.Group:
    nx_instrument = nx_entry.create_group("NXinstrument")
    nx_instrument.create_dataset("proton_charge", data=data['proton_charge'].values)

    nx_detector_1 = nx_instrument.create_group("detector_1")
    # Detector counts
    _create_compressed_dataset(
        root_entry=nx_detector_1,
        name="counts",
        var=data['counts'],
    )
    # Time of arrival bin edges
    _create_dataset_from_var(
        root_entry=nx_detector_1,
        var=data['counts'].coords['t'],
        name="t_bin",
        long_name="t_bin TOF (ms)",
    )
    # Pixel IDs
    _create_compressed_dataset(
        root_entry=nx_detector_1,
        name="pixel_id",
        var=data['counts'].coords['id'],
        long_name="pixel ID",
    )
    return nx_instrument


def _create_detector_group(data: sc.DataGroup, nx_entry: h5py.Group) -> h5py.Group:
    nx_detector = nx_entry.create_group("NXdetector")
    # Position of the first pixel (lowest ID) in the detector
    _create_compressed_dataset(
        root_entry=nx_detector,
        name="origin",
        var=data['origin_position'],
    )
    # Fast axis, along where the pixel ID increases by 1
    _create_dataset_from_var(
        root_entry=nx_detector, var=data['fast_axis'], name="fast_axis"
    )
    # Slow axis, along where the pixel ID increases
    # by the number of pixels in the fast axis
    _create_dataset_from_var(
        root_entry=nx_detector, var=data['slow_axis'], name="slow_axis"
    )
    return nx_detector


def _create_source_group(data: sc.DataGroup, nx_entry: h5py.Group) -> h5py.Group:
    nx_source = nx_entry.create_group("NXsource")
    nx_source["name"] = "European Spallation Source"
    nx_source["short_name"] = "ESS"
    nx_source["type"] = "Spallation Neutron Source"
    nx_source["distance"] = sc.norm(data['source_position']).value
    nx_source["probe"] = "neutron"
    nx_source["target_material"] = "W"
    return nx_source


def export_as_nexus(
    data: sc.DataGroup, output_file: str | pathlib.Path | io.BytesIO
) -> None:
    """Export the reduced data to a NeXus file.

    Currently exporting step is not expected to be part of sciline pipelines.
    """
    warnings.warn(
        DeprecationWarning(
            "Exporting to custom NeXus format will be deprecated in the near future "
            ">=26.12.0. "
            "Please use ``export_as_nxlauetof`` instead."
        ),
        stacklevel=2,
    )
    with h5py.File(output_file, "w") as f:
        f.attrs["default"] = "NMX_data"
        nx_entry = _create_root_data_entry(f)
        _create_sample_group(data, nx_entry)
        _create_instrument_group(data, nx_entry)
        _create_detector_group(data, nx_entry)
        _create_source_group(data, nx_entry)


def _create_lauetof_data_entry(file_obj: h5py.File) -> h5py.Group:
    nx_entry = file_obj.create_group("entry")
    nx_entry.attrs["NX_class"] = "NXentry"
    return nx_entry


def _add_lauetof_definition(nx_entry: h5py.Group) -> None:
    _create_dataset_from_string(root_entry=nx_entry, name="definition", var="NXlauetof")


def _add_lauetof_instrument(nx_entry: h5py.Group) -> h5py.Group:
    nx_instrument = nx_entry.create_group("instrument")
    nx_instrument.attrs["NX_class"] = "NXinstrument"
    _create_dataset_from_string(root_entry=nx_instrument, name="name", var="NMX")
    return nx_instrument


def _add_lauetof_source_group(
    dg: NMXExperimentMetadata, nx_instrument: h5py.Group
) -> None:
    nx_source = nx_instrument.create_group("source")
    nx_source.attrs["NX_class"] = "NXsource"
    _create_dataset_from_string(
        root_entry=nx_source, name="name", var="European Spallation Source"
    )
    _create_dataset_from_string(root_entry=nx_source, name="short_name", var="ESS")
    _create_dataset_from_string(
        root_entry=nx_source, name="type", var="Spallation Neutron Source"
    )
    _create_dataset_from_var(
        root_entry=nx_source, name="distance", var=sc.norm(dg["source_position"])
    )
    # Legacy probe information.
    _create_dataset_from_string(root_entry=nx_source, name="probe", var="neutron")


def _add_lauetof_detector_group(dg: sc.DataGroup, nx_instrument: h5py.Group) -> None:
    nx_detector = nx_instrument.create_group(dg["detector_name"].value)  # Detector name
    nx_detector.attrs["NX_class"] = "NXdetector"
    _create_dataset_from_var(
        name="polar_angle",
        root_entry=nx_detector,
        var=sc.scalar(0, unit='deg'),  # TODO: Add real data
    )
    _create_dataset_from_var(
        name="azimuthal_angle",
        root_entry=nx_detector,
        var=sc.scalar(0, unit='deg'),  # TODO: Add real data
    )
    _create_dataset_from_var(
        name="x_pixel_size", root_entry=nx_detector, var=dg["x_pixel_size"]
    )
    _create_dataset_from_var(
        name="y_pixel_size", root_entry=nx_detector, var=dg["y_pixel_size"]
    )
    _create_dataset_from_var(
        name="distance",
        root_entry=nx_detector,
        var=sc.scalar(0, unit='m'),  # TODO: Add real data
    )
    # Legacy geometry information until we have a better way to store it
    _create_dataset_from_var(
        name="origin", root_entry=nx_detector, var=dg['origin_position']
    )
    # Fast axis, along where the pixel ID increases by 1
    _create_dataset_from_var(
        root_entry=nx_detector, var=dg['fast_axis'], name="fast_axis"
    )
    # Slow axis, along where the pixel ID increases
    # by the number of pixels in the fast axis
    _create_dataset_from_var(
        root_entry=nx_detector, var=dg['slow_axis'], name="slow_axis"
    )


def _add_lauetof_sample_group(dg: NMXExperimentMetadata, nx_entry: h5py.Group) -> None:
    nx_sample = nx_entry.create_group("sample")
    nx_sample.attrs["NX_class"] = "NXsample"
    _create_dataset_from_var(
        root_entry=nx_sample,
        var=dg['crystal_rotation'],
        name='crystal_rotation',
        long_name='crystal rotation in Phi (XYZ)',
    )
    _create_dataset_from_string(
        root_entry=nx_sample,
        name='name',
        var=dg['sample_name'].value,
    )
    _create_dataset_from_var(
        name='orientation_matrix',
        root_entry=nx_sample,
        var=sc.array(
            dims=['i', 'j'],
            values=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
            unit="dimensionless",
        ),  # TODO: Add real data, the sample orientation matrix
    )
    _create_dataset_from_var(
        name='unit_cell',
        root_entry=nx_sample,
        var=sc.array(
            dims=['i'],
            values=[1.0, 1.0, 1.0, 90.0, 90.0, 90.0],
            unit="dimensionless",  # TODO: Add real data,
            # a, b, c, alpha, beta, gamma
        ),
    )


def _add_lauetof_monitor_group(data: sc.DataGroup, nx_entry: h5py.Group) -> None:
    nx_monitor = nx_entry.create_group("control")
    nx_monitor.attrs["NX_class"] = "NXmonitor"
    _create_dataset_from_string(root_entry=nx_monitor, name='mode', var='monitor')
    nx_monitor["preset"] = 0.0  # Check if this is the correct value
    data_dset = _create_dataset_from_var(
        name='data',
        root_entry=nx_monitor,
        var=sc.array(
            dims=['tof'], values=[1, 1, 1], unit="counts"
        ),  # TODO: Add real data, bin values
    )
    data_dset.attrs["signal"] = 1
    data_dset.attrs["primary"] = 1
    _create_dataset_from_var(
        name='time_of_flight',
        root_entry=nx_monitor,
        var=sc.array(
            dims=['tof'], values=[1, 1, 1], unit="s"
        ),  # TODO: Add real data, bin edges
    )


def _add_arbitrary_metadata(
    nx_entry: h5py.Group, **arbitrary_metadata: sc.Variable
) -> None:
    if not arbitrary_metadata:
        return

    metadata_group = nx_entry.create_group("metadata")
    for key, value in arbitrary_metadata.items():
        if not isinstance(value, sc.Variable):
            import warnings

            msg = f"Skipping metadata key '{key}' as it is not a scipp.Variable."
            warnings.warn(UserWarning(msg), stacklevel=2)
            continue
        else:
            _create_dataset_from_var(
                name=key,
                root_entry=metadata_group,
                var=value,
            )


def _export_static_metadata_as_nxlauetof(
    experiment_metadata: NMXExperimentMetadata,
    output_file: str | pathlib.Path | io.BytesIO,
    **arbitrary_metadata: sc.Variable,
) -> None:
    """Export the metadata to a NeXus file with the LAUE_TOF application definition.

    ``Metadata`` in this context refers to the information
    that is not part of the reduced detector counts itself,
    but is necessary for the interpretation of the reduced data.
    Since NMX can have arbitrary number of detectors,
    this function can take multiple detector metadata objects.

    Parameters
    ----------
    experiment_metadata:
        Experiment metadata object.
    output_file:
        Output file path.
    arbitrary_metadata:
        Arbitrary metadata that does not fit into the existing metadata objects.

    """
    with h5py.File(output_file, "w") as f:
        f.attrs["NX_class"] = "NXlauetof"
        nx_entry = _create_lauetof_data_entry(f)
        _add_lauetof_definition(nx_entry)
        _add_lauetof_sample_group(experiment_metadata, nx_entry)
        nx_instrument = _add_lauetof_instrument(nx_entry)
        _add_lauetof_source_group(experiment_metadata, nx_instrument)
        # Placeholder for ``monitor`` group
        _add_lauetof_monitor_group(experiment_metadata, nx_entry)
        # Skipping ``NXdata``(name) field with data link
        # Add arbitrary metadata
        _add_arbitrary_metadata(nx_entry, **arbitrary_metadata)


def _export_detector_metadata_as_nxlauetof(
    *detector_metadatas: NMXDetectorMetadata,
    output_file: str | pathlib.Path | io.BytesIO,
    append_mode: bool = True,
) -> None:
    """Export the detector specific metadata to a NeXus file.

    Since NMX can have arbitrary number of detectors,
    this function can take multiple detector metadata objects.

    Parameters
    ----------
    detector_metadatas:
        Detector metadata objects.
    output_file:
        Output file path.

    """

    if not append_mode:
        raise NotImplementedError("Only append mode is supported for now.")

    with h5py.File(output_file, "r+") as f:
        nx_entry = f["entry"]
        if "instrument" not in nx_entry:
            nx_instrument = _add_lauetof_instrument(f["entry"])
        else:
            nx_instrument = nx_entry["instrument"]
        # Add detector group metadata
        for detector_metadata in detector_metadatas:
            _add_lauetof_detector_group(detector_metadata, nx_instrument)


def _extract_counts(dg: sc.DataGroup) -> sc.Variable:
    counts: sc.DataArray = dg['counts'].data
    if 'id' in counts.dims:
        num_x, num_y = dg["detector_shape"].value
        return sc.fold(counts, dim='id', sizes={'x': num_x, 'y': num_y})
    else:
        # If there is no 'id' dimension, we assume it is already in the correct shape
        return counts


def _export_reduced_data_as_nxlauetof(
    dg: NMXReducedDataGroup,
    output_file: str | pathlib.Path | io.BytesIO,
    *,
    append_mode: bool = True,
    compress_counts: bool = True,
) -> None:
    """Export the reduced data to a NeXus file with the LAUE_TOF application definition.

    Even though this function only exports
    reduced data(detector counts and its coordinates),
    the input should contain all the necessary metadata
    for minimum sanity check.

    Parameters
    ----------
    dg:
        Reduced data and metadata.
    output_file:
        Output file path.
    append_mode:
        If ``True``, the file is opened in append mode.
        If ``False``, the file is opened in None-append mode.
        > None-append mode is not supported for now.
        > Only append mode is supported for now.
    compress_counts:
        If ``True``, the detector counts are compressed using bitshuffle.
        It is because only the detector counts are expected to be large.

    """
    if not append_mode:
        raise NotImplementedError("Only append mode is supported for now.")

    with h5py.File(output_file, "r+") as f:
        nx_detector: h5py.Group = f[f"entry/instrument/{dg['detector_name'].value}"]
        # Data - shape: [n_x_pixels, n_y_pixels, n_tof_bins]
        # The actual application definition defines it as integer,
        # but we keep the original data type for now
        num_x, num_y = dg["detector_shape"].value  # Probably better way to do this
        if compress_counts:
            data_dset = _create_compressed_dataset(
                name="data",
                root_entry=nx_detector,
                var=_extract_counts(dg),
                chunks=(num_x, num_y, 1),
                dtype=np.uint,
            )
        else:
            data_dset = _create_dataset_from_var(
                name="data",
                root_entry=nx_detector,
                var=_extract_counts(dg),
                dtype=np.uint,
            )
        data_dset.attrs["signal"] = 1
        _create_dataset_from_var(
            name='time_of_flight',
            root_entry=nx_detector,
            var=sc.midpoints(dg['counts'].coords['t'], dim='t'),
        )


def _check_file(
    filename: str | pathlib.Path | io.BytesIO, overwrite: bool
) -> pathlib.Path | io.BytesIO:
    if isinstance(filename, str | pathlib.Path):
        filename = pathlib.Path(filename)
        if filename.exists() and not overwrite:
            raise FileExistsError(
                f"File '{filename}' already exists. Use `overwrite=True` to overwrite."
            )
    return filename


T = TypeVar("T", bound=sc.DataArray)


class NXLauetofWriter:
    def __init__(
        self,
        *,
        output_filename: str | pathlib.Path | io.BytesIO,
        workflow: sl.Pipeline,
        chunk_generator: Callable[[FilePath, DetectorName], Generator[T, None, None]],
        chunk_insert_key: type[T],
        extra_meta: dict[str, sc.Variable] | None = None,
        compress_counts: bool = True,
        overwrite: bool = False,
    ) -> None:
        from ess.reduce.streaming import EternalAccumulator, StreamProcessor

        from .types import FilePath, NMXReducedCounts

        self.compress_counts = compress_counts
        self._chunk_generator = chunk_generator
        self._chunk_insert_key = chunk_insert_key
        self._workflow = workflow
        self._output_filename = _check_file(output_filename, overwrite)
        self._input_filename = workflow.compute(FilePath)
        self._final_stream_processor = partial(
            StreamProcessor,
            dynamic_keys=(chunk_insert_key,),
            target_keys=(NMXReducedDataGroup,),
            accumulators={NMXReducedCounts: EternalAccumulator},
        )
        self._detector_metas: dict[DetectorName, NMXDetectorMetadata] = {}
        self._detector_reduced: dict[DetectorName, NMXReducedDataGroup] = {}
        _export_static_metadata_as_nxlauetof(
            experiment_metadata=self._workflow.compute(NMXExperimentMetadata),
            output_file=self._output_filename,
            **(extra_meta or {}),
        )

    def add_panel(
        self, *, detector_id: DetectorIndex | DetectorName
    ) -> NMXReducedDataGroup:
        from .types import PixelIds

        temp_wf = self._workflow.copy()
        if isinstance(detector_id, int):
            temp_wf[DetectorIndex] = detector_id
        elif isinstance(detector_id, str):
            temp_wf[DetectorName] = detector_id
        else:
            raise TypeError(
                f"Expected detector_id to be an int or str, got {type(detector_id)}"
            )

        _export_detector_metadata_as_nxlauetof(
            temp_wf.compute(NMXDetectorMetadata),
            output_file=self._output_filename,
        )
        # First compute static information
        detector_name = temp_wf.compute(DetectorName)
        temp_wf[PixelIds] = temp_wf.compute(PixelIds)
        processor = self._final_stream_processor(temp_wf)
        # Then iterate over the chunks
        for da in self._chunk_generator(self._input_filename, detector_name):
            if any(da.sizes.values()) == 0:
                continue
            else:
                results = processor.add_chunk({self._chunk_insert_key: da})

        _export_reduced_data_as_nxlauetof(
            results[NMXReducedDataGroup],
            self._output_filename,
            compress_counts=self.compress_counts,
        )
        return results[NMXReducedDataGroup]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import pathlib

import pytest
import scipp as sc

from ess.nmx.data import get_small_mcstas
from ess.nmx.mcstas.load import load_raw_event_data, raw_event_data_chunk_generator


@pytest.fixture(params=[get_small_mcstas])
def mcstas_file_path(request: pytest.FixtureRequest) -> pathlib.Path:
    return request.param()


def test_generator_loading_at_once(mcstas_file_path) -> None:
    from ess.nmx.mcstas.load import detector_name_from_index

    detector_name = detector_name_from_index(0)
    whole_chunk = next(
        raw_event_data_chunk_generator(
            mcstas_file_path, detector_name=detector_name, chunk_size=-1
        )
    )
    loaded_data = load_raw_event_data(
        mcstas_file_path, detector_name=detector_name, bank_prefix=None
    )
    assert sc.identical(whole_chunk, loaded_data)


def test_generator_loading_warns_if_too_small(mcstas_file_path) -> None:
    from ess.nmx.mcstas.load import detector_name_from_index

    detector_name = detector_name_from_index(0)
    with pytest.warns(UserWarning, match="The chunk size may be too small"):
        next(
            raw_event_data_chunk_generator(
                mcstas_file_path, detector_name=detector_name, chunk_size=1
            )
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from typing import Any

import scipp as sc
import scippnexus as snx
from ess.reduce.streaming import Accumulator

from .load import _validate_chunk_size, load_event_data_bank_name
from .types import DetectorBankPrefix, DetectorName, FilePath


class MinAccumulator(Accumulator):
    """Accumulator that keeps track of the maximum value seen so far."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._cur_min: sc.Variable | None = None

    @property
    def value(self) -> sc.Variable | None:
        return self._cur_min

    def _do_push(self, value: sc.Variable) -> None:
        new_min = value.min()
        if self._cur_min is None:
            self._cur_min = new_min
        else:
            self._cur_min = min(self._cur_min, new_min)


class MaxAccumulator(Accumulator):
    """Accumulator that keeps track of the maximum value seen so far."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._cur_max: sc.Variable | None = None

    @property
    def value(self) -> sc.Variable | None:
        return self._cur_max

    def _do_push(self, value: sc.Variable) -> None:
        new_max = value.max()
        if self._cur_max is None:
            self._cur_max = new_max
        else:
            self._cur_max = max(self._cur_max, new_max)


def calculate_number_of_chunks(
    file_path: FilePath,
    *,
    detector_name: DetectorName,
    bank_prefix: DetectorBankPrefix | None = None,
    chunk_size: int = 0,  # Number of rows to read at a time
) -> int:
    """Calculate number of chunks in the event data.

    Parameters
    ----------
    file_path:
        Path to the nexus file
    detector_name:
        Name of the detector to load
    pixel_ids:
        Pixel ids to generate the data array with the events
    chunk_size:
        Number of rows to read at a time.
        If 0, chunk slice is determined automatically by the ``iter_chunks``.
        Note that it only works if the dataset is already chunked.

    Returns
    -------
    :
        Number of chunks in the event data.

    Raises
    ------
    ValueError:
        If the chunk size is not valid. (>= -1)
    TypeError:
        If the chunk size is not an integer.

    """
    _validate_chunk_size(chunk_size)
    # Find the data bank name associated with the detector
    bank_prefix = load_event_data_bank_name(
        detector_name=detector_name, file_path=file_path
    )
    bank_name = f'{bank_prefix}_dat_list_p_x_y_n_id_t'
    with snx.File(file_path, 'r') as f:
        root = f["entry1/data"]
        (bank_name,) = (name for name in root.keys() if bank_name in name)
    with snx.File(file_path, 'r') as f:
        root = f["entry1/data"]
        dset: snx.Field = root[bank_name]["events"]
        if chunk_size == 0:
            return len(list(dset.dataset.iter_chunks()))
        elif chunk_size == -1:
            return 1  # Read all at once
        else:
            return dset.shape[0] // chunk_size + int(dset.shape[0] % chunk_size != 0)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2026 Scipp contributors (https://github.com/scipp)
from dataclasses import fields, is_dataclass

import scipp as sc


def _is_nested(obj) -> bool:
    return is_dataclass(obj) or isinstance(obj, sc.DataGroup | dict)


def to_datagroup(obj) -> sc.DataGroup:
    if is_dataclass(obj):
        return sc.DataGroup(
            {
                field.name: to_datagroup(value)
                if _is_nested(value := getattr(obj, field.name))
                else value
                for field in fields(obj)
            }
        )
    elif isinstance(obj, sc.DataGroup | dict):
        return sc.DataGroup(
            {
                name: to_datagroup(value) if _is_nested(value) else value
                for name, value in obj.items()
            }
        )
    else:
        return obj



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# McStas instrument geometry xml description related functions.
from collections.abc import Iterable
from dataclasses import dataclass
from types import MappingProxyType
from typing import Protocol, TypeVar

import h5py
import scipp as sc
from defusedxml.ElementTree import fromstring

from ..rotation import axis_angle_to_quaternion, quaternion_to_matrix
from .types import FilePath

T = TypeVar('T')


_AXISNAME_TO_UNIT_VECTOR = MappingProxyType(
    {
        'x': sc.vector([1.0, 0.0, 0.0]),
        'y': sc.vector([0.0, 1.0, 0.0]),
        'z': sc.vector([0.0, 0.0, 1.0]),
    }
)


class _XML(Protocol):
    """XML element or tree type.

    Temporarily used for type hinting.
    Builtin XML type is blocked by bandit security check."""

    tag: str
    attrib: dict[str, str]

    def find(self, name: str) -> '_XML | None': ...

    def __iter__(self) -> '_XML': ...

    def __next__(self) -> '_XML': ...


def _check_and_unpack_if_only_one(xml_items: list[_XML], name: str) -> _XML:
    """Check if there is only one element with ``name``."""
    if len(xml_items) > 1:
        raise ValueError(f"Multiple {name}s found.")
    elif len(xml_items) == 0:
        raise ValueError(f"No {name} found.")

    return xml_items.pop()


def select_by_tag(xml_items: _XML, tag: str) -> _XML:
    """Select element with ``tag`` if there is only one."""

    return _check_and_unpack_if_only_one(list(filter_by_tag(xml_items, tag)), tag)


def filter_by_tag(xml_items: Iterable[_XML], tag: str) -> Iterable[_XML]:
    """Filter xml items by tag."""
    return (item for item in xml_items if item.tag == tag)


def filter_by_type_prefix(xml_items: Iterable[_XML], prefix: str) -> Iterable[_XML]:
    """Filter xml items by type prefix."""
    return (
        item for item in xml_items if item.attrib.get('type', '').startswith(prefix)
    )


def select_by_type_prefix(xml_items: Iterable[_XML], prefix: str) -> _XML:
    """Select xml item by type prefix."""

    cands = list(filter_by_type_prefix(xml_items, prefix))
    return _check_and_unpack_if_only_one(cands, prefix)


def find_attributes(component: _XML, *args: str) -> dict[str, float]:
    """Retrieve ``args`` as float from xml."""

    return {key: float(component.attrib[key]) for key in args}


@dataclass
class SimulationSettings:
    """Simulation settings extracted from McStas instrument xml description."""

    # From <defaults>
    length_unit: str  # 'unit' of <length>
    angle_unit: str  # 'unit' of <angle>
    # From <reference-frame>
    beam_axis: str  # 'axis' of <along-beam>
    handedness: str  # 'val' of <handedness>

    @classmethod
    def from_xml(cls, tree: _XML) -> 'SimulationSettings':
        """Create simulation settings from xml."""
        defaults = select_by_tag(tree, 'defaults')
        length_desc = select_by_tag(defaults, 'length')
        angle_desc = select_by_tag(defaults, 'angle')
        reference_frame = select_by_tag(defaults, 'reference-frame')
        along_beam = select_by_tag(reference_frame, 'along-beam')
        handedness = select_by_tag(reference_frame, 'handedness')

        return cls(
            length_unit=length_desc.attrib['unit'],
            angle_unit=angle_desc.attrib['unit'],
            beam_axis=along_beam.attrib['axis'],
            handedness=handedness.attrib['val'],
        )


def _position_from_location(location: _XML, unit: str = 'm') -> sc.Variable:
    """Retrieve position from location."""
    x, y, z = find_attributes(location, 'x', 'y', 'z').values()
    return sc.vector([x, y, z], unit=unit)


def _rotation_matrix_from_location(
    location: _XML, angle_unit: str = 'degree'
) -> sc.Variable:
    """Retrieve rotation matrix from location."""

    attribs = find_attributes(location, 'axis-x', 'axis-y', 'axis-z', 'rot')
    x, y, z, w = axis_angle_to_quaternion(
        x=attribs['axis-x'],
        y=attribs['axis-y'],
        z=attribs['axis-z'],
        theta=sc.scalar(-attribs['rot'], unit=angle_unit),
    )
    return quaternion_to_matrix(x=x, y=y, z=z, w=w)


@dataclass
class DetectorDesc:
    """Detector information extracted from McStas instrument xml description."""

    # From <component type="MonNDtype-n" ...>
    component_type: str  # 'type'
    name: str
    id_start: int  # 'idstart'
    fast_axis_name: str  # 'idfillbyfirst'
    # From <type name="MonNDtype-n" ...>
    num_x: int  # 'xpixels'
    num_y: int  # 'ypixels'
    step_x: sc.Variable  # 'xstep'
    step_y: sc.Variable  # 'ystep'
    start_x: float  # 'xstart'
    start_y: float  # 'ystart'
    # From <location> under <component type="MonNDtype-n" ...>
    position: sc.Variable  # <location> 'x', 'y', 'z'
    # Calculated fields
    rotation_matrix: sc.Variable
    slow_axis_name: str
    fast_axis: sc.Variable
    slow_axis: sc.Variable

    @classmethod
    def from_xml(
        cls,
        *,
        component: _XML,
        type_desc: _XML,
        simulation_settings: SimulationSettings,
    ) -> 'DetectorDesc':
        """Create detector description from xml component and type."""

        location = select_by_tag(component, 'location')
        rotation_matrix = _rotation_matrix_from_location(
            location, simulation_settings.angle_unit
        )
        fast_axis_name = component.attrib['idfillbyfirst']
        slow_axis_name = 'xy'.replace(fast_axis_name, '')

        length_unit = simulation_settings.length_unit

        # Type casting from str to float and then int to allow *e* notation
        # For example, '1e4' -> 10000.0 -> 10_000
        return cls(
            component_type=type_desc.attrib['name'],
            name=component.attrib['name'],
            id_start=int(float(component.attrib['idstart'])),
            fast_axis_name=fast_axis_name,
            slow_axis_name=slow_axis_name,
            num_x=int(float(type_desc.attrib['xpixels'])),
            num_y=int(float(type_desc.attrib['ypixels'])),
            step_x=sc.scalar(float(type_desc.attrib['xstep']), unit=length_unit),
            step_y=sc.scalar(float(type_desc.attrib['ystep']), unit=length_unit),
            start_x=float(type_desc.attrib['xstart']),
            start_y=float(type_desc.attrib['ystart']),
            position=_position_from_location(location, simulation_settings.length_unit),
            rotation_matrix=rotation_matrix,
            fast_axis=rotation_matrix * _AXISNAME_TO_UNIT_VECTOR[fast_axis_name],
            slow_axis=rotation_matrix * _AXISNAME_TO_UNIT_VECTOR[slow_axis_name],
        )

    @property
    def total_pixels(self) -> int:
        return self.num_x * self.num_y

    @property
    def slow_step(self) -> sc.Variable:
        return self.step_y if self.fast_axis_name == 'x' else self.step_x

    @property
    def fast_step(self) -> sc.Variable:
        return self.step_x if self.fast_axis_name == 'x' else self.step_y

    @property
    def num_fast_pixels_per_row(self) -> int:
        """Number of pixels in each row of the detector along the fast axis."""
        return self.num_x if self.fast_axis_name == 'x' else self.num_y

    @property
    def detector_shape(self) -> tuple:
        """Shape of the detector panel. (num_x, num_y)"""
        return (self.num_x, self.num_y)


def _collect_detector_descriptions(tree: _XML) -> tuple[DetectorDesc, ...]:
    """Retrieve detector geometry descriptions from mcstas file."""
    type_list = list(filter_by_tag(tree, 'type'))
    simulation_settings = SimulationSettings.from_xml(tree)

    def _find_type_desc(det: _XML) -> _XML:
        for type_ in type_list:
            if type_.attrib['name'] == det.attrib['type']:
                return type_

        raise ValueError(
            f"Cannot find type {det.attrib['type']} for {det.attrib['name']}."
        )

    detector_components = [
        DetectorDesc.from_xml(
            component=det,
            type_desc=_find_type_desc(det),
            simulation_settings=simulation_settings,
        )
        for det in filter_by_type_prefix(filter_by_tag(tree, 'component'), 'MonNDtype')
    ]

    return tuple(sorted(detector_components, key=lambda x: x.id_start))


@dataclass
class SampleDesc:
    """Sample description extracted from McStas instrument xml description."""

    # From <component type="sampleMantid-type" ...>
    component_type: str
    name: str
    # From <location> under <component type="sampleMantid-type" ...>
    position: sc.Variable
    rotation_matrix: sc.Variable | None

    @classmethod
    def from_xml(
        cls, *, tree: _XML, simulation_settings: SimulationSettings
    ) -> 'SampleDesc':
        """Create sample description from xml component."""
        source_xml = select_by_type_prefix(tree, 'sampleMantid-type')
        location = select_by_tag(source_xml, 'location')
        try:
            rotation_matrix = _rotation_matrix_from_location(
                location, simulation_settings.angle_unit
            )
        except KeyError:
            rotation_matrix = None

        return cls(
            component_type=source_xml.attrib['type'],
            name=source_xml.attrib['name'],
            position=_position_from_location(location, simulation_settings.length_unit),
            rotation_matrix=rotation_matrix,
        )

    def position_from_sample(self, other: sc.Variable) -> sc.Variable:
        """Position of ``other`` relative to the sample.

        All positions and distance are stored relative to the sample position.

        Parameters
        ----------
        other:
            Position of the other object in 3D vector.

        """
        return other - self.position


@dataclass
class SourceDesc:
    """Source description extracted from McStas instrument xml description."""

    # From <component type="Source" ...>
    component_type: str
    name: str
    # From <location> under <component type="Source" ...>
    position: sc.Variable

    @classmethod
    def from_xml(
        cls, *, tree: _XML, simulation_settings: SimulationSettings
    ) -> 'SourceDesc':
        """Create source description from xml component."""
        source_xml = select_by_type_prefix(tree, 'sourceMantid-type')
        location = select_by_tag(source_xml, 'location')

        return cls(
            component_type=source_xml.attrib['type'],
            name=source_xml.attrib['name'],
            position=_position_from_location(location, simulation_settings.length_unit),
        )


def _construct_pixel_id(detector_desc: DetectorDesc) -> sc.Variable:
    """Pixel IDs for single detector."""
    start, stop = (
        detector_desc.id_start,
        detector_desc.id_start + detector_desc.total_pixels,
    )
    return sc.arange('id', start, stop, unit=None)


def _construct_pixel_ids(detector_descs: tuple[DetectorDesc, ...]) -> sc.Variable:
    """Pixel IDs for all detectors."""
    ids = [_construct_pixel_id(det) for det in detector_descs]
    return sc.concat(ids, 'id')


def _pixel_positions(
    detector: DetectorDesc, position_offset: sc.Variable
) -> sc.Variable:
    """Position of pixels of the ``detector``.

    Position of each pixel is relative to the position_offset.
    """
    pixel_idx = sc.arange('id', detector.total_pixels)
    n_col = sc.scalar(detector.num_fast_pixels_per_row)

    pixel_n_slow = pixel_idx // n_col
    pixel_n_fast = pixel_idx % n_col

    fast_axis_steps = detector.fast_axis * detector.fast_step
    slow_axis_steps = detector.slow_axis * detector.slow_step

    return (
        (pixel_n_slow * slow_axis_steps)
        + (pixel_n_fast * fast_axis_steps)
        + detector.rotation_matrix
        * sc.vector(
            [detector.start_x, detector.start_y, 0.0], unit=position_offset.unit
        )  # Detector pixel offset should also be rotated first.
    ) + position_offset


@dataclass
class McStasInstrument:
    simulation_settings: SimulationSettings
    detectors: tuple[DetectorDesc, ...]
    source: SourceDesc
    sample: SampleDesc

    @classmethod
    def from_xml(cls, tree: _XML) -> 'McStasInstrument':
        """Create McStas instrument from xml."""
        simulation_settings = SimulationSettings.from_xml(tree)

        return cls(
            simulation_settings=simulation_settings,
            detectors=_collect_detector_descriptions(tree),
            source=SourceDesc.from_xml(
                tree=tree, simulation_settings=simulation_settings
            ),
            sample=SampleDesc.from_xml(
                tree=tree, simulation_settings=simulation_settings
            ),
        )

    def pixel_ids(self, *det_names: str) -> sc.Variable:
        """Pixel IDs for the detectors.

        If multiple detectors are requested, all pixel IDs will be concatenated along
        the 'id' dimension.

        Parameters
        ----------
        det_names:
            Names of the detectors to extract pixel IDs for.

        """
        detectors = tuple(det for det in self.detectors if det.name in det_names)
        return _construct_pixel_ids(detectors)

    def experiment_metadata(self) -> dict[str, sc.Variable]:
        """Extract experiment metadata from the McStas instrument description."""
        return {
            'sample_position': self.sample.position_from_sample(self.sample.position),
            'source_position': self.sample.position_from_sample(self.source.position),
            'sample_name': sc.scalar(self.sample.name),
        }

    def _detector_metadata(self, det_name: str) -> dict[str, sc.Variable]:
        try:
            detector = next(det for det in self.detectors if det.name == det_name)
        except StopIteration as e:
            raise KeyError(f"Detector {det_name} not found.") from e
        return {
            'fast_axis': detector.fast_axis,
            'slow_axis': detector.slow_axis,
            'origin_position': self.sample.position_from_sample(detector.position),
            'position': _pixel_positions(
                detector, self.sample.position_from_sample(detector.position)
            ),
            'detector_shape': sc.scalar(detector.detector_shape),
            'x_pixel_size': detector.step_x,
            'y_pixel_size': detector.step_y,
            'detector_name': sc.scalar(detector.name),
        }

    def detector_metadata(self, *det_names: str) -> dict[str, sc.Variable]:
        """Extract detector metadata from the McStas instrument description.

        If multiple detector is requested, all metadata will be concatenated along the
        'panel' dimension.

        Parameters
        ----------
        det_names:
            Names of the detectors to extract metadata for.

        """
        if len(det_names) == 1:
            return self._detector_metadata(det_names[0])
        detector_metadatas = {
            det_name: self._detector_metadata(det_name) for det_name in det_names
        }
        # Concat all metadata into panel dimension
        metadata_keys: set[str] = set().union(
            set(detector_metadatas[det_name].keys()) for det_name in det_names
        )
        return {
            key: sc.concat(
                [metadata[key] for metadata in detector_metadatas.values()], 'panel'
            )
            for key in metadata_keys
        }


def read_mcstas_geometry_xml(file_path: FilePath) -> McStasInstrument:
    """Retrieve geometry parameters from mcstas file"""
    instrument_xml_path = 'entry1/instrument/instrument_xml/data'
    with h5py.File(file_path) as file:
        tree = fromstring(file[instrument_xml_path][...][0])
        return McStasInstrument.from_xml(tree)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import enum

from pydantic import BaseModel, Field

from .types import Compression


class InputConfig(BaseModel):
    # Add title of the basemodel
    model_config = {"title": "Input Configuration"}
    # File IO
    input_file: list[str] = Field(
        title="Input File",
        description="Path to the input file. If multiple file paths are given,"
        " the output(histogram) will be merged(summed) "
        "and will not save individual outputs per input file. ",
    )
    swmr: bool = Field(
        title="SWMR Mode",
        description="Open the input file in SWMR mode",
        default=False,
    )
    # Detector selection
    detector_ids: list[int] = Field(
        title="Detector IDs",
        description="Detector indices to process",
        default=[0, 1, 2],
    )
    # Chunking options
    iter_chunk: bool = Field(
        title="Iterate in Chunks",
        description="Whether to process the input file in chunks "
        " based on the hdf5 dataset chunk size. "
        "It is ignored if hdf5 dataset is not chunked. "
        "If True, it overrides chunk-size-pulse and chunk-size-events options.",
        default=False,
    )
    chunk_size_pulse: int = Field(
        title="Chunk Size Pulse",
        description="Number of pulses to process in each chunk. "
        "If 0 or negative, process all pulses at once.",
        default=0,
    )
    chunk_size_events: int = Field(
        title="Chunk Size Events",
        description="Number of events to process in each chunk. "
        "If 0 or negative, process all events at once."
        "If both chunk-size-pulse and chunk-size-events are set, "
        "chunk-size-pulse is preferred.",
        default=0,
    )


class TimeBinUnit(enum.StrEnum):
    ms = 'ms'
    us = 'us'
    ns = 'ns'


class TimeBinCoordinate(enum.StrEnum):
    event_time_offset = 'event_time_offset'
    time_of_flight = 'time_of_flight'


class WorkflowConfig(BaseModel):
    # Add title of the basemodel
    model_config = {"title": "Workflow Configuration"}
    time_bin_coordinate: TimeBinCoordinate = Field(
        title="Time Bin Coordinate",
        description="Coordinate to bin the time data. "
        "Selecting `event_time_offset` means "
        "reduction steps are skipped, "
        "i.e. calculating `time of flight(tof)` "
        "and simply saves histograms of the raw data.",
        default=TimeBinCoordinate.time_of_flight,
        # Default is time of flight since
        # DIALS should expect the time of flight.
    )
    nbins: int = Field(
        title="Number of Time Bins",
        description="Number of Time bins",
        default=50,
    )
    min_time_bin: int | None = Field(
        title="Minimum Time",
        description="Minimum time edge of [time_bin_coordinate] in [time_bin_unit].",
        default=None,
    )
    max_time_bin: int | None = Field(
        title="Maximum Time",
        description="Maximum time edge of [time_bin_coordinate] in [time_bin_unit].",
        default=None,
    )
    time_bin_unit: TimeBinUnit = Field(
        title="Unit of Time Bins",
        description="Unit of time bins.",
        default=TimeBinUnit.ms,
    )
    tof_lookup_table_file_path: str | None = Field(
        title="TOF Lookup Table File Path",
        description="Path to the TOF lookup table file. "
        "If None, the lookup table will be computed on-the-fly.",
        default=None,
    )
    tof_simulation_num_neutrons: int = Field(
        title="Number of Neutrons for TOF Simulation",
        description="Number of neutrons to simulate for TOF lookup table calculation.",
        default=1_000_000,
    )
    tof_simulation_min_wavelength: float = Field(
        title="TOF Simulation Minimum Wavelength",
        description="Minimum wavelength for TOF simulation in Angstrom.",
        default=1.8,
    )
    tof_simulation_max_wavelength: float = Field(
        title="TOF Simulation Maximum Wavelength",
        description="Maximum wavelength for TOF simulation in Angstrom.",
        default=3.6,
    )
    tof_simulation_min_ltotal: float = Field(
        title="TOF Simulation Minimum Ltotal",
        description="Minimum total flight path for TOF simulation in meters.",
        default=150.0,
    )
    tof_simulation_max_ltotal: float = Field(
        title="TOF Simulation Maximum Ltotal",
        description="Maximum total flight path for TOF simulation in meters.",
        default=170.0,
    )
    tof_simulation_seed: int = Field(
        title="TOF Simulation Seed",
        description="Random seed for TOF simulation.",
        default=42,  # No reason.
    )


class OutputConfig(BaseModel):
    # Add title of the basemodel
    model_config = {"title": "Output Configuration"}
    # Log verbosity
    verbose: bool = Field(
        title="Verbose Logging",
        description="Increase output verbosity.",
        default=False,
    )
    # File output
    skip_file_output: bool = Field(
        title="Skip File Output",
        description="If True, the output file will not be written.",
        default=False,
    )
    output_file: str = Field(
        title="Output File",
        description="Path to the output file. "
        "It will be overwritten if ``overwrite`` is True.",
        default="scipp_output.h5",
    )
    overwrite: bool = Field(
        title="Overwrite Output File",
        description="If True, overwrite the output file if ``output_file`` exists.",
        default=False,
    )
    compression: Compression = Field(
        title="Compression",
        description="Compress option of reduced output file.",
        default=Compression.BITSHUFFLE_LZ4,
    )


class ReductionConfig(BaseModel):
    """Container for all reduction configurations."""

    inputs: InputConfig
    workflow: WorkflowConfig = Field(default_factory=WorkflowConfig)
    output: OutputConfig = Field(default_factory=OutputConfig)

    @property
    def _children(self) -> list[BaseModel]:
        return [self.inputs, self.workflow, self.output]


def to_command_arguments(
    *, config: ReductionConfig, one_line: bool = True, separator: str = '\\\n'
) -> list[str] | str:
    """Convert the config to a list of command line arguments.

    Parameters
    ----------
    one_line:
        If True, return a single string with all arguments joined by spaces.
        If False, return a list of argument strings.

    """
    args = {}
    for instance in config._children:
        args.update(instance.model_dump(mode='python'))
    args = {f"--{k.replace('_', '-')}": v for k, v in args.items() if v is not None}

    arg_list = []
    for k, v in args.items():
        if not isinstance(v, bool):
            arg_list.append(k)
            if isinstance(v, list):
                arg_list.extend(str(item) for item in v)
            elif isinstance(v, enum.StrEnum):
                arg_list.append(v.value)
            else:
                arg_list.append(str(v))
        elif v is True:
            arg_list.append(k)

    if one_line:
        # Default separator is backslash + newline for better readability
        # Users can directly copy-paste the output in a terminal or a script.
        return (
            (separator + '--')
            .join(" ".join(arg_list).split('--'))
            .removeprefix(separator)
        )
    else:
        return arg_list



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# Rotation related functions for NMX
import numpy as np
import scipp as sc
from numpy.typing import NDArray


def axis_angle_to_quaternion(
    *, x: float, y: float, z: float, theta: sc.Variable
) -> NDArray:
    """Convert axis-angle to queternions, [x, y, z, w].

    Parameters
    ----------
    x:
        X component of axis of rotation.
    y:
        Y component of axis of rotation.
    z:
        Z component of axis of rotation.
    theta:
        Angle of rotation, with unit of ``rad`` or ``deg``.

    Returns
    -------
    :
        A list of (normalized) quaternions, [x, y, z, w].

    Notes
    -----
    Axis of rotation (x, y, z) does not need to be normalized,
    but it returns a unit quaternion (x, y, z, w).

    """

    w: sc.Variable = sc.cos(theta.to(unit='rad') / 2)
    xyz: sc.Variable = -sc.sin(theta.to(unit='rad') / 2) * sc.vector([x, y, z])
    q = np.array([*xyz.values, w.value])
    return q / np.linalg.norm(q)


def quaternion_to_matrix(*, x: float, y: float, z: float, w: float) -> sc.Variable:
    """Convert quaternion to rotation matrix.

    Parameters
    ----------
    x:
        x(a) component of quaternion.
    y:
        y(b) component of quaternion.
    z:
        z(c) component of quaternion.
    w:
        w component of quaternion.

    Returns
    -------
    :
        A 3x3 rotation matrix.

    """
    from scipy.spatial.transform import Rotation

    return sc.spatial.rotations_from_rotvecs(
        rotation_vectors=sc.vector(
            Rotation.from_quat([x, y, z, w]).as_rotvec(),
            unit='rad',
        )
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pathlib

import gemmi
import pytest
import scipp as sc

from ess.nmx import mtz_io
from ess.nmx.data import get_small_mtz_samples
from ess.nmx.mtz_io import (
    DEFAULT_SPACE_GROUP_DESC,  # P 1
    MtzDataFrame,
    MTZFileIndex,
    MTZFilePath,
    NMXMtzDataArray,
    NMXMtzDataFrame,
    get_reciprocal_asu,
    mtz_to_pandas,
    nmx_mtz_dataframe_to_scipp_dataarray,
    process_mtz_dataframe,
    process_single_mtz_to_dataframe,
    read_mtz_file,
)


@pytest.fixture(params=get_small_mtz_samples())
def file_path(request) -> pathlib.Path:
    return request.param


def test_gemmi_mtz(file_path: pathlib.Path) -> None:
    mtz = read_mtz_file(MTZFilePath(file_path))
    assert mtz.spacegroup == gemmi.SpaceGroup("C 1 2 1")  # Hard-coded value
    assert len(mtz.columns[0]) == 100  # Number of samples, hard-coded value


@pytest.fixture
def gemmi_mtz_object(file_path: pathlib.Path) -> gemmi.Mtz:
    return read_mtz_file(MTZFilePath(file_path))


def test_mtz_to_pandas_dataframe(gemmi_mtz_object: gemmi.Mtz) -> None:
    df = mtz_to_pandas(gemmi_mtz_object)
    assert set(df.columns) == set(gemmi_mtz_object.column_labels())
    # Check if the test data are not all-same
    first_column_name, second_column_name = df.columns[0:2]
    assert not all(df[first_column_name] == df[second_column_name])

    # Check if the data are the same
    for column in gemmi_mtz_object.columns:
        assert column.label in df.columns
        assert all(df[column.label] == column.array)


def test_mtz_to_process_pandas_dataframe(gemmi_mtz_object: gemmi.Mtz) -> None:
    df = process_single_mtz_to_dataframe(gemmi_mtz_object)
    for expected_colum in ["hkl", "d", "resolution", *"HKL", "wavelength", "I", "SIGI"]:
        assert expected_colum in df.columns

    for hkl_column in "HKL":
        assert hkl_column in df.columns
        assert df[hkl_column].dtype == int

    assert "hkl_asu" not in df.columns  # It should be done on merged dataframes


@pytest.fixture
def mtz_list() -> list[gemmi.Mtz]:
    return [
        read_mtz_file(MTZFilePath(file_path)) for file_path in get_small_mtz_samples()
    ]


def test_get_space_group_with_spacegroup_desc() -> None:
    assert (
        mtz_io.get_space_group_from_description(DEFAULT_SPACE_GROUP_DESC).short_name()
        == "P1"
    )


@pytest.fixture
def conflicting_mtz_series(
    mtz_list: list[gemmi.Mtz],
) -> list[gemmi.Mtz]:
    mtz_list[MTZFileIndex(0)].spacegroup = gemmi.SpaceGroup(DEFAULT_SPACE_GROUP_DESC)
    # Make sure the space groups are different
    assert (
        mtz_list[MTZFileIndex(0)].spacegroup.short_name()
        != mtz_list[MTZFileIndex(1)].spacegroup.short_name()
    )

    return mtz_list


def test_get_unique_space_group_raises_on_conflict(
    conflicting_mtz_series: list[gemmi.Mtz],
) -> None:
    reg = r"Multiple space groups found:.+P 1.+C 1 2 1"
    space_groups = [
        mtz_io.get_space_group_from_mtz(mtz) for mtz in conflicting_mtz_series
    ]
    with pytest.raises(ValueError, match=reg):
        mtz_io.get_unique_space_group(*space_groups)


@pytest.fixture
def merged_mtz_dataframe(mtz_list: list[gemmi.Mtz]) -> MtzDataFrame:
    """Tests if the merged data frame has the expected columns."""
    reduced_mtz = [process_single_mtz_to_dataframe(mtz) for mtz in mtz_list]
    return mtz_io.merge_mtz_dataframes(*reduced_mtz)


@pytest.fixture
def nmx_data_frame(
    mtz_list: list[gemmi.Mtz],
    merged_mtz_dataframe: MtzDataFrame,
) -> NMXMtzDataFrame:
    space_grs = [mtz_io.get_space_group_from_mtz(mtz) for mtz in mtz_list]
    space_gr = mtz_io.get_unique_space_group(*space_grs)
    reciprocal_asu = get_reciprocal_asu(space_gr)

    return process_mtz_dataframe(
        mtz_df=merged_mtz_dataframe,
        reciprocal_asu=reciprocal_asu,
        sg=space_gr,
    )


def test_process_merged_mtz_dataframe(
    merged_mtz_dataframe: MtzDataFrame,
    nmx_data_frame: NMXMtzDataFrame,
) -> None:
    assert "hkl_asu" not in merged_mtz_dataframe.columns
    assert "hkl_asu" in nmx_data_frame.columns


@pytest.fixture
def nmx_data_array(nmx_data_frame: NMXMtzDataFrame) -> NMXMtzDataArray:
    return nmx_mtz_dataframe_to_scipp_dataarray(nmx_data_frame)


def test_to_scipp_dataarray(
    nmx_data_array: NMXMtzDataArray,
) -> None:
    # Check the intended modification
    for indices_coord_name in ("hkl", "hkl_asu"):
        assert nmx_data_array.coords[indices_coord_name].dtype == str

    assert sc.all(nmx_data_array.data > 0)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import json
import pathlib
from typing import NewType

import gemmi
import numpy as np
import pandas as pd
import scipp as sc

from .dials_reflection_io import load

# User defined or configurable types
DialsReflectionFilePath = NewType("DialsReflectionFilePath", pathlib.Path)
"""Path to the dials reflection file"""
DialsReflectionFile = NewType("DialsReflectionFile", dict)
"""The raw DIALS reflection file, read in as a dict"""
DialsExperimentFilePath = NewType("DialsExperimentFilePath", pathlib.Path)
"""Path to the dials experiment file"""
DialsExperiment = NewType("DialsExperiment", dict)
"""Experiment details from DIALS .expt file (JSON format)"""
SpaceGroupDesc = NewType("SpaceGroupDesc", str)
"""The space group description. e.g. 'P 21 21 21'"""
DEFAULT_SPACE_GROUP_DESC = SpaceGroupDesc("P 1")
"""The default space group description to use if not found in the input files."""
UnitCell = NewType("UnitCell", tuple[float])
"""The unit cell a, b, c in Angstrom, alpha, beta, gamma in degrees"""

# Custom column names
WavelengthColumnName = NewType("WavelengthColumnName", str)
"""The name of the wavelength column in the DIALS reflection file."""
DEFAULT_WAVELENGTH_COLUMN_NAME = WavelengthColumnName("LAMBDA")

IntensityColumnName = NewType("IntensityColumnName", str)
"""The name of the intensity column in the DIALS reflection file."""
DEFAULT_INTENSITY_COLUMN_NAME = IntensityColumnName("I")

VarianceColumnName = NewType("VarianceColumnName", str)
"""The name of the variance (stdev(I)**2) of intensity column in the DIALS reflection
file."""
DEFAULT_VARIANCE_COLUMN_NAME = VarianceColumnName("VARI")

StdDevColumnName = NewType("StdDevColumnName", str)
"""The name of the standard deviation of intensity column in the DIALS reflection
file."""
DEFAULT_STDEV_COLUMN_NAME = VarianceColumnName("SIGI")


# Computed types
DialsDataFrame = NewType("DialsDataFrame", pd.DataFrame)
"""The raw mtz dataframe."""
NMXDialsDataFrame = NewType("NMXDialsDataFrame", pd.DataFrame)
"""The processed mtz dataframe with derived columns."""
NMXDialsDataArray = NewType("NMXDialsDataArray", sc.DataArray)


def read_dials_reflection_file(
    file_path: DialsReflectionFilePath,
) -> DialsReflectionFile:
    """read dials reflection file"""

    return DialsReflectionFile(load(file_path.as_posix(), copy=True))


def read_dials_experiment_file(file_path: DialsExperimentFilePath) -> DialsExperiment:
    """Read Dials Experiment .expt file"""

    return DialsExperiment(json.load(open(file_path)))


def get_unit_cell(dials_expt: DialsExperiment) -> UnitCell:
    """
    Get the unit cell from the expt file.
    It is saved as real-space vectors so the unit cell has to be
    calculated from them.
    """
    crystal = dials_expt['crystal'][0]
    ra, rb, rc = tuple([crystal[f'real_space_{x}'] for x in 'abc'])
    a = np.linalg.norm(ra)
    b = np.linalg.norm(rb)
    c = np.linalg.norm(rc)
    al = np.rad2deg(np.arccos(np.dot(rb, rc) / (b * c)))
    be = np.rad2deg(np.arccos(np.dot(ra, rc) / (a * c)))
    ga = np.rad2deg(np.arccos(np.dot(ra, rb) / (a * b)))

    return UnitCell(a, b, c, al, be, ga)


def get_unique_space_group(dials_expt: DialsExperiment) -> gemmi.SpaceGroup:
    """
    Get space group from Dials expt file.
    For cctbx/disambiguation reasons it is saved as the Hall symbol, but
    the H-M notation can be back-determined with gemmi.
    """
    crystal = dials_expt['crystal'][0]
    sg_hall = crystal['space_group_hall_symbol']

    return gemmi.find_spacegroup_by_ops(gemmi.symops_from_hall(sg_hall))


def get_reciprocal_asu(spacegroup: gemmi.SpaceGroup) -> gemmi.ReciprocalAsu:
    """Returns the reciprocal asymmetric unit from the space group."""

    return gemmi.ReciprocalAsu(spacegroup)


def dials_refl_to_pandas(refls: dict) -> pd.DataFrame:
    """Converts the loaded DIALS reflection file to a pandas dataframe.

    It is equivalent to the following code:

    .. code-block:: python

        import numpy as np
        import pandas as pd

        data = np.array(mtz, copy=False)
        columns = mtz.column_labels()
        return pd.DataFrame(data, columns=columns)

    It is recommended in the gemmi documentation.

    """
    if refls.get('experiment_identifier'):  # this has no relevant information
        del refls['experiment_identifier']  # and it complicates loading as a df
    return pd.DataFrame(
        {
            key: list(val) if isinstance(val, np.ndarray) and val.ndim > 1 else val
            for key, val in refls.items()
        }
    )


def process_dials_refl_list_to_dataframe(
    refls: dict,
) -> DialsDataFrame:
    """Select and derive columns from the original ``MtzDataFrame``.

    Parameters
    ----------
    mtz:
        The raw mtz dataset.

    wavelength_column_name:
        The name of the wavelength column in the mtz file.

    intensity_column_name:
        The name of the intensity column in the mtz file.

    intensity_sig_col_name:
        The name of the standard uncertainty of intensity column in the mtz file.

    Returns
    -------
    :
        The new mtz dataframe with derived and renamed columns.

        The derived columns are:

        - ``SIGI``: The uncertainty of the intensity value, defined as the square root
        of the measured variance.

        For consistent names of columns/coordinates, the following columns are renamed:

        - ``wavelength_column_name`` -> ``'wavelength'``
        - ``intensity_column_name`` -> ``'I'``
        - ``intensity_sig_col_name`` -> ``'SIGI'``

        Other columns are kept as they are.

    """
    orig_df = dials_refl_to_pandas(refls)
    new_df = pd.DataFrame()

    new_df['H'] = orig_df['miller_index'].map(lambda x: x[0]).astype(int)
    new_df['K'] = orig_df['miller_index'].map(lambda x: x[1]).astype(int)
    new_df['L'] = orig_df['miller_index'].map(lambda x: x[2]).astype(int)

    new_df['hkl'] = orig_df['miller_index']
    new_df["d"] = orig_df['d']
    new_df['wavelength'] = orig_df['wavelength_cal']

    new_df[DEFAULT_INTENSITY_COLUMN_NAME] = orig_df['intensity.sum.value']
    new_df[DEFAULT_VARIANCE_COLUMN_NAME] = orig_df['intensity.sum.variance']
    new_df[DEFAULT_STDEV_COLUMN_NAME] = np.sqrt(orig_df['intensity.sum.variance'])

    for column in [col for col in orig_df.columns if col not in new_df]:
        new_df[column] = orig_df[column]

    return DialsDataFrame(new_df)


def process_dials_dataframe(
    *,
    dials_df: DialsDataFrame,
    reciprocal_asu: gemmi.ReciprocalAsu,
    sg: gemmi.SpaceGroup,
) -> NMXDialsDataFrame:
    """Modify/Add columns of the shallow copy of a dials dataframe.

    This method must be called after merging multiple mtz dataframe.
    """

    df = dials_df.copy(deep=False)

    def _reciprocal_asu(row: pd.Series) -> list[int]:
        """Converts miller indices(HKL) to ASU indices."""

        return reciprocal_asu.to_asu(row["hkl"], sg.operations())[0]

    df["hkl_asu"] = df.apply(_reciprocal_asu, axis=1)
    # Unpack the indices for later.
    df[["H_ASU", "K_ASU", "L_ASU"]] = pd.DataFrame(
        df["hkl_asu"].to_list(), index=df.index
    )

    return NMXDialsDataFrame(df)


def nmx_dials_dataframe_to_scipp_dataarray(
    nmx_mtz_df: NMXDialsDataFrame,
) -> NMXDialsDataArray:
    """Converts the processed mtz dataframe to a scipp dataarray.

    The intensity, with column name :attr:`~DEFAULT_INTENSITY_COLUMN_NAME`
    becomes the data and the standard uncertainty of intensity,
    with column name :attr:`~DEFAULT_SIGMA_INTENSITY_COLUMN_NAME`
    becomes the variances of the data.

    Parameters
    ----------
    nmx_mtz_df:
        The merged and processed mtz dataframe.

    Returns
    -------
    :
        The scipp dataarray with the intensity and variances.
        The ``I`` column becomes the data and the
        squared ``SIGI`` column becomes the variances.
        Therefore they are not in the coordinates.

        Following coordinates are modified:

        - ``hkl``: The miller indices as a string.
                   It is modified to have a string dtype
                   since is no dtype that can represent this in scipp.

        - ``hkl_asu``: The asymmetric unit of miller indices as a string.
                       This coordinate will be used to derive estimated scale factors.
                       It is modified to have a string dtype
                       as the same reason as why ``hkl`` coordinate is modified.

        Zero or negative intensities are removed from the dataarray.
        It can happen due to the post-processing of the data,
        e.g. background subtraction.

    """
    from scipp.compat.pandas_compat import from_pandas_dataframe, parse_bracket_header

    to_scipp = nmx_mtz_df.copy(deep=False)
    # Convert to scipp Dataset
    nmx_mtz_ds = from_pandas_dataframe(
        to_scipp,
        data_columns=[
            DEFAULT_INTENSITY_COLUMN_NAME,
            DEFAULT_STDEV_COLUMN_NAME,
            DEFAULT_VARIANCE_COLUMN_NAME,
        ],
        header_parser=parse_bracket_header,
    )
    # Pop the indices columns.
    # TODO: We can put them back once we support tuple[int] dtype.
    # See https://github.com/scipp/scipp/issues/3046 for more details.
    # Temporarily, we will manually convert them to a string.
    # It is done on the scipp variable instead of the dataframe
    # since columns with string dtype are converted to PyObject dtype
    # instead of string by `from_pandas_dataframe`.
    for indices_name in ("hkl", "hkl_asu"):
        nmx_mtz_ds.coords[indices_name] = sc.array(
            dims=nmx_mtz_ds.coords[indices_name].dims,
            values=nmx_mtz_df[indices_name].astype(str).tolist(),
            # `astype`` is not enough to convert the dtype to string.
            # The result of `astype` will have `PyObject` as a dtype.
        )
    # Add units
    nmx_mtz_ds.coords["wavelength"].unit = sc.units.angstrom
    for key in nmx_mtz_ds.keys():
        nmx_mtz_ds[key].unit = sc.units.dimensionless

    # Add variances
    nmx_mtz_da = nmx_mtz_ds[DEFAULT_INTENSITY_COLUMN_NAME].copy(deep=False)
    nmx_mtz_da.variances = nmx_mtz_ds[DEFAULT_VARIANCE_COLUMN_NAME].values

    # Return DataArray without negative intensities
    return NMXDialsDataArray(nmx_mtz_da[nmx_mtz_da.data > 0])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import logging
import pathlib
import warnings
from collections.abc import Callable

import numpy as np
import scipp as sc
import scippnexus as snx
from ess.reduce.nexus.types import Filename, NeXusName, RawDetector, SampleRun
from ess.reduce.time_of_flight.types import TimeOfFlightLookupTable, TofDetector

from ._executable_helper import (
    build_logger,
    build_reduction_argument_parser,
    collect_matching_input_files,
    reduction_config_from_args,
)
from .configurations import (
    OutputConfig,
    ReductionConfig,
    TimeBinCoordinate,
    WorkflowConfig,
)
from .nexus import (
    _check_file,
    export_detector_metadata_as_nxlauetof,
    export_monitor_metadata_as_nxlauetof,
    export_reduced_data_as_nxlauetof,
    export_static_metadata_as_nxlauetof,
)
from .types import (
    NMXDetectorMetadata,
    NMXInstrument,
    NMXLauetof,
    NMXMonitorMetadata,
    NMXReducedDetector,
    NMXSampleMetadata,
    NMXSourceMetadata,
)
from .workflows import initialize_nmx_workflow, select_detector_names

_TOF_COORD_NAME = 'tof'
"""Name of the TOF coordinate used in DataArrays."""
_ETO_COORD_NAME = 'event_time_offset'
"""Name of the Event Time Offset Coordinate used in Nexus."""


def _retrieve_input_file(input_file: list[str]) -> pathlib.Path:
    """Temporary helper to retrieve a single input file from the list
    Until multiple input file support is implemented.
    """
    from collections import Counter

    # Check duplicated pattern or paths
    _counts = Counter(input_file)
    duplicating_patterns = {pattern for pattern, num in _counts.items() if num > 1}
    if duplicating_patterns:
        raise ValueError(
            f"Duplicated file paths or pattern found. {duplicating_patterns}"
        )

    if isinstance(input_file, list):
        input_files = collect_matching_input_files(*input_file)
        if len(input_files) == 0:
            raise ValueError(
                "No input files found for reduction."
                "Check if the file paths are correct.",
                input_file,
            )
        elif len(input_files) > 1:
            raise NotImplementedError(
                "Currently, only a single input file is supported for reduction."
            )
        input_file_path = input_files[0]
    else:
        input_file_path = input_file

    return input_file_path


def _retrieve_display(
    logger: logging.Logger | None, display: Callable | None
) -> Callable:
    if display is not None:
        return display
    elif logger is not None:
        return logger.info
    else:
        return logging.getLogger(__name__).info


def _retrieve_time_bin_coordinate_name(wf_config: WorkflowConfig) -> str:
    if wf_config.time_bin_coordinate == TimeBinCoordinate.time_of_flight:
        return _TOF_COORD_NAME
    elif wf_config.time_bin_coordinate == TimeBinCoordinate.event_time_offset:
        return _ETO_COORD_NAME


def _warn_bin_edge_out_of_range(
    *, edge: sc.Variable, coord_name: str, desc: str
) -> None:
    warnings.warn(
        message=f"{edge} is {desc} than all "
        f"{coord_name} values.\n"
        "The histogram will all have zero values.",
        category=UserWarning,
        stacklevel=4,
    )


def _match_data_unit_dtype(config_var: sc.Variable, da: sc.Variable) -> sc.Variable:
    return config_var.to(unit=da.unit, dtype=da.dtype)


def _build_time_bin_edges(
    *,
    wf_config: WorkflowConfig,
    result_das: sc.DataGroup,
    t_coord_name: str,
) -> sc.Variable:
    # Calculate the min and max of the data itself.
    da_min_t = min(da.bins.coords[t_coord_name].nanmin() for da in result_das.values())
    da_max_t = max(da.bins.coords[t_coord_name].nanmax() for da in result_das.values())

    # Use the user-set parameters if available
    # and validate them according to the data.
    # Lower Time Bin Edge
    if wf_config.min_time_bin is not None:
        min_t = sc.scalar(wf_config.min_time_bin, unit=wf_config.time_bin_unit)
        min_t = _match_data_unit_dtype(min_t, da=da_min_t)
        # If the user-set minimum time bin value
        # is bigger than all time-bin-coordinate values.
        if min_t > da_max_t:
            _warn_bin_edge_out_of_range(
                edge=min_t, coord_name=wf_config.time_bin_coordinate, desc='bigger'
            )
    else:
        min_t = da_min_t

    # Upper Time Bin Edge
    if wf_config.max_time_bin is not None:
        max_t = sc.scalar(wf_config.max_time_bin, unit=wf_config.time_bin_unit)
        max_t = _match_data_unit_dtype(max_t, da=da_max_t)
        # If the user-set maximum time bin value
        # is smaller than all time-bin-coordinate values.
        if max_t <= da_min_t:
            _warn_bin_edge_out_of_range(
                edge=max_t, coord_name=wf_config.time_bin_coordinate, desc='smaller'
            )
    else:
        max_t = da_max_t

    # Avoid dropping the event that has the exact same
    # `event_time_offset`` or `tof` value as the upper bin edge.
    max_t.value = np.nextafter(max_t.value, np.inf)

    # Validate the results.
    if min_t >= max_t:
        raise ValueError(
            f"Minimum time bin edge, {min_t} "
            "is bigger than or equal to the "
            f"maximum time bin edge, {max_t}.\n"
            "Cannot build a time bin edges coordinate.\n"
            "Please check your configurations again."
        )

    # Build the bin-edges to histogram the results.
    n_edges = wf_config.nbins + 1
    return sc.linspace(dim=t_coord_name, start=min_t, stop=max_t, num=n_edges)


def reduction(
    *,
    config: ReductionConfig,
    logger: logging.Logger | None = None,
    display: Callable | None = None,
) -> NMXLauetof:
    """Reduce NMX data from a Nexus file and export to NXLauetof(ESS NMX specific) file.

    Parameters
    ----------
    config:
        Reduction configuration.

        Data reduction parameters are taken from this config
        instead of passing them directly as keyword arguments.
        They can be either built from command-line arguments
        using `ReductionConfig.from_args()` or constructed manually.

        If the reduced data is successfully written to the output file
        the configuration is also saved there for future reference.
    logger:
        Logger to use for logging messages. If None, a default logger is created.
    display:
        Callable for displaying messages, useful in Jupyter notebooks. If None,
        defaults to logger.info.

    Returns
    -------
    sc.DataGroup:
        A DataGroup containing the reduced data for each selected detector.

    """
    # Check the file output configuration before we start heavy computation.
    if not config.output.skip_file_output:
        _check_file(config.output.output_file, config.output.overwrite)

    display = _retrieve_display(logger, display)
    input_file_path = _retrieve_input_file(config.inputs.input_file).resolve()
    display(f"Input file: {input_file_path}")

    output_file_path = pathlib.Path(config.output.output_file).resolve()
    display(f"Output file: {output_file_path}")

    detector_names = select_detector_names(detector_ids=config.inputs.detector_ids)

    # Initialize workflow
    base_wf = initialize_nmx_workflow(config=config.workflow)
    # Insert parameters and cache intermediate results
    base_wf[Filename[SampleRun]] = input_file_path

    if config.workflow.time_bin_coordinate == TimeBinCoordinate.time_of_flight:
        # We cache the time of flight look up table
        # only if we need to calculate time-of-flight coordinates.
        # If `event_time_offset` was requested,
        # we do not have to calculate the look up table at all.
        base_wf[TimeOfFlightLookupTable] = base_wf.compute(TimeOfFlightLookupTable)

    metadatas = base_wf.compute((NMXSampleMetadata, NMXSourceMetadata))

    tof_das = sc.DataGroup()
    detector_metas = sc.DataGroup()

    if config.workflow.time_bin_coordinate == TimeBinCoordinate.event_time_offset:
        target_type = RawDetector[SampleRun]
    elif config.workflow.time_bin_coordinate == TimeBinCoordinate.time_of_flight:
        target_type = TofDetector[SampleRun]

    for detector_name in detector_names:
        cur_wf = base_wf.copy()
        cur_wf[NeXusName[snx.NXdetector]] = detector_name
        results = cur_wf.compute((target_type, NMXDetectorMetadata))
        detector_metas[detector_name] = results[NMXDetectorMetadata]
        # Binning into 1 bin and getting final tof bin edges later.
        tof_das[detector_name] = results[target_type]

    # Make tof bin edges covering all detectors
    t_coord_name = _retrieve_time_bin_coordinate_name(wf_config=config.workflow)
    t_bin_edges = _build_time_bin_edges(
        wf_config=config.workflow, result_das=tof_das, t_coord_name=t_coord_name
    )

    monitor_metadata = NMXMonitorMetadata(
        tof_bin_coord=t_coord_name,
        # TODO: Use real monitor data
        # Currently NMX simulations or experiments do not have monitors
        data=sc.DataArray(
            coords={t_coord_name: t_bin_edges},
            data=sc.ones_like(t_bin_edges[:-1]),
        ),
    )

    # Histogram detector counts
    tof_histograms = sc.DataGroup()
    for detector_name, tof_da in tof_das.items():
        histogram = tof_da.hist({t_coord_name: t_bin_edges})
        tof_histograms[detector_name] = histogram

    detector_results = sc.DataGroup(
        {
            detector_name: NMXReducedDetector(
                data=histogram, metadata=detector_metas[detector_name]
            )
            for detector_name, histogram in tof_histograms.items()
        }
    )
    source_meta: NMXSourceMetadata = metadatas[NMXSourceMetadata]
    sample_meta: NMXSampleMetadata = metadatas[NMXSampleMetadata]

    results = NMXLauetof(
        control=monitor_metadata,
        instrument=NMXInstrument(detectors=detector_results, source=source_meta),
        sample=sample_meta,
    )

    if config.workflow.time_bin_coordinate == TimeBinCoordinate.time_of_flight:
        results.lookup_table = base_wf.compute(TimeOfFlightLookupTable)

    if not config.output.skip_file_output:
        save_results(results=results, output_config=config.output)

    return results


def save_results(*, results: NMXLauetof, output_config: OutputConfig) -> None:
    # Validate if results have expected fields

    export_static_metadata_as_nxlauetof(
        sample_metadata=results.sample,
        source_metadata=results.instrument.source,
        program=results.reducer,
        output_file=output_config.output_file,
        overwrite=output_config.overwrite,
    )
    export_monitor_metadata_as_nxlauetof(
        monitor_metadata=results.control,
        output_file=output_config.output_file,
    )
    for detector_name, detector_result in results.instrument.detectors.items():
        export_detector_metadata_as_nxlauetof(
            detector_metadata=detector_result.metadata,
            output_file=output_config.output_file,
        )
        if isinstance(detector_result.data, sc.DataArray):
            export_reduced_data_as_nxlauetof(
                detector_name=detector_name,
                da=detector_result.data,
                output_file=output_config.output_file,
                compress_mode=output_config.compression,
            )
        else:
            raise ValueError(f"Detector counts histogram missing in {detector_name}")


def main() -> None:
    parser = build_reduction_argument_parser()
    config = reduction_config_from_args(parser.parse_args())
    logger = build_logger(config.output)

    reduction(config=config, logger=logger)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from collections.abc import Iterable

import sciline
import scipp as sc
import scippnexus as snx
import tof
from ess.reduce.nexus.types import (
    EmptyDetector,
    Filename,
    NeXusComponent,
    NeXusTransformation,
    Position,
    SampleRun,
)
from ess.reduce.time_of_flight import (
    GenericTofWorkflow,
    LookupTableRelativeErrorThreshold,
    LtotalRange,
    NumberOfSimulatedNeutrons,
    SimulationResults,
    SimulationSeed,
    TofLookupTableWorkflow,
)
from ess.reduce.time_of_flight.lut import BeamlineComponentReading
from ess.reduce.time_of_flight.types import TimeOfFlightLookupTableFilename
from ess.reduce.workflow import register_workflow

from .configurations import WorkflowConfig
from .types import (
    NMXDetectorMetadata,
    NMXSampleMetadata,
    NMXSourceMetadata,
    TofSimulationMaxWavelength,
    TofSimulationMinWavelength,
)

default_parameters = {
    TofSimulationMaxWavelength: sc.scalar(3.6, unit='angstrom'),
    TofSimulationMinWavelength: sc.scalar(1.8, unit='angstrom'),
    LookupTableRelativeErrorThreshold: {f'detector_panel_{i}': 0.1 for i in range(5)},
}


def _simulate_fixed_wavelength_tof(
    wmin: TofSimulationMinWavelength,
    wmax: TofSimulationMaxWavelength,
    neutrons: NumberOfSimulatedNeutrons,
    seed: SimulationSeed,
) -> SimulationResults:
    """
    Simulate a pulse of neutrons propagating through the instrument using the
    ``tof`` package (https://tof.readthedocs.io).
    This runs a simulation assuming there are no choppers in the instrument.

    Parameters
    ----------
    wmin:
        Minimum wavelength of the simulated neutrons.
    wmax:
        Maximum wavelength of the simulated neutrons.
    neutrons:
        Number of neutrons to simulate.
    seed:
        Random seed for the simulation.
    """
    source = tof.Source(
        facility="ess",
        neutrons=neutrons,
        pulses=1,
        seed=seed,
        wmax=wmax,
        wmin=wmin,
    )
    events = source.data.squeeze().flatten(to="event")

    return SimulationResults(
        readings={
            "source": BeamlineComponentReading(
                time_of_arrival=events.coords["birth_time"],
                wavelength=events.coords["wavelength"],
                weight=events.data,
                distance=source.distance,
            )
        },
        choppers=None,
    )


def _merge_panels(*da: sc.DataArray) -> sc.DataArray:
    """Merge multiple DataArrays representing different panels into one."""
    merged = sc.concat(da, dim='panel')
    return merged


def select_detector_names(*, detector_ids: Iterable[int] = (0, 1, 2)):
    import os

    # Users can override detector names via environment variable
    # It is a comma-separated list of detector names
    # e.g., NMX_DETECTOR_NAMES=detector_panel_0,detector_panel_1,detector_panel_2
    # The detector names are not expected to be changed from the default ones,
    # but this option is provided for minimum flexibility.
    DETECTOR_NAME_VAR = os.environ.get("NMX_DETECTOR_NAMES", None)
    if DETECTOR_NAME_VAR is not None:
        return tuple(
            name
            for i_name, name in enumerate(DETECTOR_NAME_VAR.split(','))
            if i_name in detector_ids
        )
    else:
        return tuple(f'detector_panel_{i}' for i in detector_ids)


def assemble_sample_metadata(
    crystal_rotation: Position[snx.NXcrystal, SampleRun],
    sample_position: Position[snx.NXsample, SampleRun],
    sample_component: NeXusComponent[snx.NXsample, SampleRun],
) -> NMXSampleMetadata:
    """Assemble sample metadata for NMX reduction workflow."""
    name = sample_component['name']
    if isinstance(name, sc.Variable) and name.dtype == str:
        sample_name = name.value
    elif isinstance(name, str):
        sample_name = name
    else:
        raise TypeError(f'Sample name {name}is in a wrong type: ', type(name))

    return NMXSampleMetadata(
        name=sample_name,
        crystal_rotation=crystal_rotation,
        position=sample_position,
    )


def assemble_source_metadata(
    source_position: Position[snx.NXsource, SampleRun],
) -> NMXSourceMetadata:
    """Assemble source metadata for NMX reduction workflow."""
    return NMXSourceMetadata(position=source_position)


def _decide_fast_axis(da: sc.DataArray) -> str:
    x_slice = da['x_pixel_offset', 0].coords['detector_number']
    y_slice = da['y_pixel_offset', 0].coords['detector_number']

    if (x_slice.max() < y_slice.max()).value:
        return 'y'
    elif (x_slice.max() > y_slice.max()).value:
        return 'x'
    else:
        raise ValueError(
            "Cannot decide fast axis based on pixel offsets. "
            "Please specify the fast axis explicitly."
        )


def _decide_step(offsets: sc.Variable) -> sc.Variable:
    """Decide the step size based on the offsets assuming at least 2 values."""
    sorted_offsets = sc.sort(offsets, key=offsets.dim, order='ascending')
    return sorted_offsets[1] - sorted_offsets[0]


def _normalize_vector(vec: sc.Variable) -> sc.Variable:
    return vec / sc.norm(vec)


def _retrieve_crystal_rotation(
    file_path: Filename[SampleRun],
) -> Position[snx.NXcrystal, SampleRun]:
    """Temporary provider to retrieve crystal rotation from Nexus file."""
    from ess.reduce.nexus._nexus_loader import load_from_path
    from ess.reduce.nexus.types import NeXusLocationSpec

    spec = NeXusLocationSpec(
        filename=file_path,
        component_name='sample/crystal_rotation',
    )
    try:
        rotation: snx.nxtransformations.Transform = load_from_path(location=spec)
    except KeyError:
        import warnings

        warnings.warn(
            "No crystal rotation found in the Nexus file under "
            f"'entry/{spec.component_name}'. Returning zero rotation.",
            RuntimeWarning,
            stacklevel=1,
        )
        zero_rotation = sc.vector([0, 0, 0], unit='deg')
        return Position[snx.NXcrystal, SampleRun](zero_rotation)
    else:
        # TODO: Make sure if retrieving rotation vector is enough here.
        return Position[snx.NXcrystal, SampleRun](rotation.vector)


def assemble_detector_metadata(
    detector_component: NeXusComponent[snx.NXdetector, SampleRun],
    transformation: NeXusTransformation[snx.NXdetector, SampleRun],
    sample_position: Position[snx.NXsample, SampleRun],
    source_position: Position[snx.NXsource, SampleRun],
    empty_detector: EmptyDetector[SampleRun],
) -> NMXDetectorMetadata:
    """Assemble detector metadata for NMX reduction workflow."""
    positions = empty_detector.coords['position']
    # Origin should be the center of the detector.
    origin = positions.mean()
    _fast_axis = _decide_fast_axis(empty_detector)
    _slow_axis = 'y' if _fast_axis == 'x' else 'x'
    t_unit = transformation.value.unit

    axis_vectors = {
        'x': positions['x_pixel_offset', 1]['y_pixel_offset', 0]
        - positions['x_pixel_offset', 0]['y_pixel_offset', 0],
        'y': positions['y_pixel_offset', 1]['x_pixel_offset', 0]
        - positions['y_pixel_offset', 0]['x_pixel_offset', 0],
    }

    fast_axis_vector = axis_vectors[_fast_axis].to(unit=t_unit)
    slow_axis_vector = axis_vectors[_slow_axis].to(unit=t_unit)
    x_pixel_size = _decide_step(empty_detector.coords['x_pixel_offset'])
    y_pixel_size = _decide_step(empty_detector.coords['y_pixel_offset'])
    distance = sc.norm(origin - source_position.to(unit=origin.unit))

    # We save the first pixel position so that DIALS can read use it.
    flattened = empty_detector.flatten(to='detector_number')
    first_pixel_number = flattened.coords['detector_number'].min()
    first_pixel_position = flattened['detector_number', first_pixel_number].coords[
        'position'
    ]
    first_pixel_position_from_sample = first_pixel_position - sample_position

    return NMXDetectorMetadata(
        detector_name=detector_component['nexus_component_name'],
        x_pixel_size=x_pixel_size,
        y_pixel_size=y_pixel_size,
        origin=origin,
        fast_axis=_normalize_vector(fast_axis_vector),
        fast_axis_dim=_fast_axis + '_pixel_offset',
        slow_axis=_normalize_vector(slow_axis_vector),
        slow_axis_dim=_slow_axis + '_pixel_offset',
        distance=distance,
        first_pixel_position=first_pixel_position_from_sample,
    )


@register_workflow
def NMXWorkflow() -> sciline.Pipeline:
    generic_wf = GenericTofWorkflow(run_types=[SampleRun], monitor_types=[])

    generic_wf.insert(_retrieve_crystal_rotation)
    generic_wf.insert(assemble_sample_metadata)
    generic_wf.insert(assemble_source_metadata)
    generic_wf.insert(assemble_detector_metadata)
    for key, value in default_parameters.items():
        generic_wf[key] = value

    return generic_wf


def _validate_mergable_workflow(wf: sciline.Pipeline):
    if wf.indices:
        raise NotImplementedError("Only flat workflow can be merged.")


def _merge_workflows(
    base_wf: sciline.Pipeline, merged_wf: sciline.Pipeline
) -> sciline.Pipeline:
    _validate_mergable_workflow(base_wf)
    _validate_mergable_workflow(merged_wf)

    for key, spec in merged_wf.underlying_graph.nodes.items():
        if 'value' in spec:
            base_wf[key] = spec['value']
        elif (provider_spec := spec.get('provider')) is not None:
            base_wf.insert(provider_spec.func)

    return base_wf


def initialize_nmx_workflow(*, config: WorkflowConfig) -> sciline.Pipeline:
    """Initialize NMX workflow according to the workflow configuration.

    If a TOF lookup table file path is provided in the configuration,
    it is used directly. Otherwise, a TOF simulation workflow is added to
    the NMX workflow to compute the lookup table on-the-fly.

    All other parameters required for TOF simulation are also set
    as parameters in the workflow.

    Parameters
    ----------
    config:
        Workflow configuration for NMX reduction.
    params:
        Additional parameters to set in the workflow.

    """
    wf = NMXWorkflow()
    if config.tof_lookup_table_file_path is not None:
        wf[TimeOfFlightLookupTableFilename] = config.tof_lookup_table_file_path
    else:
        wf = _merge_workflows(base_wf=wf, merged_wf=TofLookupTableWorkflow())
        wf.insert(_simulate_fixed_wavelength_tof)
        wmax = sc.scalar(config.tof_simulation_max_wavelength, unit='angstrom')
        wmin = sc.scalar(config.tof_simulation_min_wavelength, unit='angstrom')
        wf[TofSimulationMaxWavelength] = wmax
        wf[TofSimulationMinWavelength] = wmin
        wf[SimulationSeed] = config.tof_simulation_seed
        ltotal_min = sc.scalar(value=config.tof_simulation_min_ltotal, unit='m')
        ltotal_max = sc.scalar(value=config.tof_simulation_max_ltotal, unit='m')
        wf[LtotalRange] = LtotalRange((ltotal_min, ltotal_max))

    return wf


__all__ = ['NMXWorkflow']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pathlib
from typing import NewType

import gemmi
import numpy as np
import pandas as pd
import scipp as sc

# Index types for param table.
MTZFileIndex = NewType("MTZFileIndex", int)
"""The index of the mtz file when iterating over multiple mtz files."""

# User defined or configurable types
MTZFilePath = NewType("MTZFilePath", pathlib.Path)
"""Path to the mtz file"""
SpaceGroupDesc = NewType("SpaceGroupDesc", str)
"""The space group description. e.g. 'P 21 21 21'"""
DEFAULT_SPACE_GROUP_DESC = SpaceGroupDesc("P 1")
"""The default space group description to use if not found in the mtz files."""

# Custom column names
WavelengthColumnName = NewType("WavelengthColumnName", str)
"""The name of the wavelength column in the mtz file."""
DEFAULT_WAVELENGTH_COLUMN_NAME = WavelengthColumnName("LAMBDA")

IntensityColumnName = NewType("IntensityColumnName", str)
"""The name of the intensity column in the mtz file."""
DEFAULT_INTENSITY_COLUMN_NAME = IntensityColumnName("I")

StdDevColumnName = NewType("StdDevColumnName", str)
"""The name of the standard uncertainty of intensity column in the mtz file."""
DEFAULT_STD_DEV_COLUMN_NAME = StdDevColumnName("SIGI")

# Computed types
MtzDataFrame = NewType("MtzDataFrame", pd.DataFrame)
"""The raw mtz dataframe."""
NMXMtzDataFrame = NewType("NMXMtzDataFrame", pd.DataFrame)
"""The processed mtz dataframe with derived columns."""
NMXMtzDataArray = NewType("NMXMtzDataArray", sc.DataArray)


def read_mtz_file(file_path: MTZFilePath) -> gemmi.Mtz:
    """read mtz file"""

    return gemmi.read_mtz_file(file_path.as_posix())


def mtz_to_pandas(mtz: gemmi.Mtz) -> pd.DataFrame:
    """Converts the mtz file to a pandas dataframe.

    It is equivalent to the following code:

    .. code-block:: python

        import numpy as np
        import pandas as pd

        data = np.array(mtz, copy=False)
        columns = mtz.column_labels()
        return pd.DataFrame(data, columns=columns)

    It is recommended in the gemmi documentation.

    """

    return pd.DataFrame(  # Recommended in the gemmi documentation.
        data=np.array(mtz, copy=False), columns=mtz.column_labels()
    )


def process_single_mtz_to_dataframe(
    mtz: gemmi.Mtz,
    wavelength_column_name: WavelengthColumnName = DEFAULT_WAVELENGTH_COLUMN_NAME,
    intensity_column_name: IntensityColumnName = DEFAULT_INTENSITY_COLUMN_NAME,
    intensity_sig_col_name: StdDevColumnName = DEFAULT_STD_DEV_COLUMN_NAME,
) -> MtzDataFrame:
    """Select and derive columns from the original ``MtzDataFrame``.

    Parameters
    ----------
    mtz:
        The raw mtz dataset.

    wavelength_column_name:
        The name of the wavelength column in the mtz file.

    intensity_column_name:
        The name of the intensity column in the mtz file.

    intensity_sig_col_name:
        The name of the standard uncertainty of intensity column in the mtz file.

    Returns
    -------
    :
        The new mtz dataframe with derived and renamed columns.

        The derived columns are:

        - ``hkl``: The miller indices as a list of integers.
        - ``d``: The d-spacing calculated from the miller indices.
                 :math:``\\dfrac{2}{d^{2}} = \\dfrac{\\sin^2(\\theta)}{\\lambda^2}``
        - ``resolution``: The resolution calculated from the d-spacing.

        For consistent names of columns/coordinates, the following columns are renamed:

        - ``wavelength_column_name`` -> ``'wavelength'``
        - ``intensity_column_name`` -> ``'I'``
        - ``intensity_sig_col_name`` -> ``'SIGI'``

        Other columns are kept as they are.

    Notes
    -----
    :class:`pandas.DataFrame` is used from loading to merging,
    but :class:`gemmi.Mtz` has :func:`gemmi.Mtz:calculate_d`
    that can derive the ``d`` using ``HKL``.
    This part of the method must be called on each mtz file separately.

    """
    from .mtz_io import mtz_to_pandas

    orig_df = mtz_to_pandas(mtz)
    mtz_df = pd.DataFrame()

    # HKL should always be integer.
    mtz_df[["H", "K", "L"]] = orig_df[["H", "K", "L"]].astype(int)
    mtz_df["hkl"] = mtz_df[["H", "K", "L"]].values.tolist()

    def _calculate_d(row: pd.Series) -> float:
        return mtz.get_cell().calculate_d(row["hkl"])

    mtz_df["d"] = mtz_df.apply(_calculate_d, axis=1)
    mtz_df["resolution"] = (1 / mtz_df["d"]) ** 2 / 4
    mtz_df["wavelength"] = orig_df[wavelength_column_name]
    mtz_df[DEFAULT_INTENSITY_COLUMN_NAME] = orig_df[intensity_column_name]
    mtz_df[DEFAULT_STD_DEV_COLUMN_NAME] = orig_df[intensity_sig_col_name]
    # Keep other columns
    for column in [col for col in orig_df.columns if col not in mtz_df]:
        mtz_df[column] = orig_df[column]

    return MtzDataFrame(mtz_df)


def get_space_group_from_description(desc: SpaceGroupDesc) -> gemmi.SpaceGroup:
    """Retrieves spacegroup from parameter.

    Parameters
    ----------
    desc:
        The space group description to use if not found in the mtz files.

    Returns
    -------
    :
        The space group.
    """
    return gemmi.SpaceGroup(desc)


def get_space_group_from_mtz(mtz: gemmi.Mtz) -> gemmi.SpaceGroup | None:
    """Retrieves spacegroup from file.

    Spacegroup is always expected in any MTZ files, but it may be missing.

    Parameters
    ----------
    mtz:
        Raw mtz dataset.

    Returns
    -------
    :
        The space group, or None if not found.
    """
    return mtz.spacegroup


def get_unique_space_group(*spacegroups: gemmi.SpaceGroup | None) -> gemmi.SpaceGroup:
    """Retrieves the unique space group from multiple space groups.

    Parameters
    ----------
    spacegroups:
        The space groups to check.

    Returns
    -------
    :
        The unique space group.

    Raises
    ------
    ValueError:
        If there are multiple space groups.
    """
    spacegroups = [sgrp for sgrp in spacegroups if sgrp is not None]
    if len(spacegroups) == 0:
        raise ValueError("No space group found.")
    first = spacegroups[0]
    if all(sgrp == first for sgrp in spacegroups):
        return first
    raise ValueError(f"Multiple space groups found: {spacegroups}")


def get_reciprocal_asu(spacegroup: gemmi.SpaceGroup) -> gemmi.ReciprocalAsu:
    """Returns the reciprocal asymmetric unit from the space group."""

    return gemmi.ReciprocalAsu(spacegroup)


def merge_mtz_dataframes(*mtz_dfs: MtzDataFrame) -> MtzDataFrame:
    """Merge multiple mtz dataframes into one."""

    return MtzDataFrame(pd.concat(mtz_dfs, ignore_index=True))


def process_mtz_dataframe(
    *,
    mtz_df: MtzDataFrame,
    reciprocal_asu: gemmi.ReciprocalAsu,
    sg: gemmi.SpaceGroup,
) -> NMXMtzDataFrame:
    """Modify/Add columns of the shallow copy of a mtz dataframe.

    This method must be called after merging multiple mtz dataframe.
    """
    df = mtz_df.copy(deep=False)

    def _reciprocal_asu(row: pd.Series) -> list[int]:
        """Converts miller indices(HKL) to ASU indices."""

        return reciprocal_asu.to_asu(row["hkl"], sg.operations())[0]

    df["hkl_asu"] = df.apply(_reciprocal_asu, axis=1)
    # Unpack the indices for later.
    df[["H_ASU", "K_ASU", "L_ASU"]] = pd.DataFrame(
        df["hkl_asu"].to_list(), index=df.index
    )

    return NMXMtzDataFrame(df)


def nmx_mtz_dataframe_to_scipp_dataarray(
    nmx_mtz_df: NMXMtzDataFrame,
) -> NMXMtzDataArray:
    """Converts the processed mtz dataframe to a scipp dataarray.

    The intensity, with column name :attr:`~DEFAULT_INTENSITY_COLUMN_NAME`
    becomes the data and the standard uncertainty of intensity,
    with column name :attr:`~DEFAULT_SIGMA_INTENSITY_COLUMN_NAME`
    becomes the variances of the data.

    Parameters
    ----------
    nmx_mtz_df:
        The merged and processed mtz dataframe.

    Returns
    -------
    :
        The scipp dataarray with the intensity and variances.
        The ``I`` column becomes the data and the
        squared ``SIGI`` column becomes the variances.
        Therefore they are not in the coordinates.

        Following coordinates are modified:

        - ``hkl``: The miller indices as a string.
                   It is modified to have a string dtype
                   since is no dtype that can represent this in scipp.

        - ``hkl_asu``: The asymmetric unit of miller indices as a string.
                       This coordinate will be used to derive estimated scale factors.
                       It is modified to have a string dtype
                       as the same reason as why ``hkl`` coordinate is modified.

        Zero or negative intensities are removed from the dataarray.
        It can happen due to the post-processing of the data,
        e.g. background subtraction.

    """
    from scipp.compat.pandas_compat import from_pandas_dataframe, parse_bracket_header

    to_scipp = nmx_mtz_df.copy(deep=False)
    # Convert to scipp Dataset
    nmx_mtz_ds = from_pandas_dataframe(
        to_scipp,
        data_columns=[
            DEFAULT_INTENSITY_COLUMN_NAME,
            DEFAULT_STD_DEV_COLUMN_NAME,
        ],
        header_parser=parse_bracket_header,
    )
    # Pop the indices columns.
    # TODO: We can put them back once we support tuple[int] dtype.
    # See https://github.com/scipp/scipp/issues/3046 for more details.
    # Temporarily, we will manually convert them to a string.
    # It is done on the scipp variable instead of the dataframe
    # since columns with string dtype are converted to PyObject dtype
    # instead of string by `from_pandas_dataframe`.
    for indices_name in ("hkl", "hkl_asu"):
        nmx_mtz_ds.coords[indices_name] = sc.array(
            dims=nmx_mtz_ds.coords[indices_name].dims,
            values=nmx_mtz_df[indices_name].astype(str).tolist(),
            # `astype`` is not enough to convert the dtype to string.
            # The result of `astype` will have `PyObject` as a dtype.
        )
    # Add units
    nmx_mtz_ds.coords["wavelength"].unit = sc.units.angstrom
    for key in nmx_mtz_ds.keys():
        nmx_mtz_ds[key].unit = sc.units.dimensionless

    # Add variances
    nmx_mtz_da = nmx_mtz_ds[DEFAULT_INTENSITY_COLUMN_NAME].copy(deep=False)
    nmx_mtz_da.variances = (nmx_mtz_ds[DEFAULT_STD_DEV_COLUMN_NAME].data ** 2).values

    # Return DataArray without negative intensities
    return NMXMtzDataArray(nmx_mtz_da[nmx_mtz_da.data > 0])


providers = (
    read_mtz_file,
    process_single_mtz_to_dataframe,
    # get_space_group_from_description,
    get_space_group_from_mtz,
    get_reciprocal_asu,
    process_mtz_dataframe,
    nmx_mtz_dataframe_to_scipp_dataarray,
)
"""The providers related to the MTZ IO."""

default_parameters = {
    WavelengthColumnName: DEFAULT_WAVELENGTH_COLUMN_NAME,
    IntensityColumnName: DEFAULT_INTENSITY_COLUMN_NAME,
    StdDevColumnName: DEFAULT_STD_DEV_COLUMN_NAME,
}
"""The parameters related to the MTZ IO."""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc

from ess.nmx.scaling import (
    ReferenceIntensities,
    estimate_scale_factor_per_hkl_asu_from_reference,
    get_reference_intensities,
    get_reference_wavelength,
)


@pytest.fixture
def nmx_data_array() -> sc.DataArray:
    da = sc.DataArray(
        data=sc.array(dims=["row"], values=[1, 2, 3, 4, 5, 3.1, 3.2]),
        coords={
            "wavelength": sc.Variable(dims=["row"], values=[1, 2, 3, 4, 5, 3, 3]),
            "hkl_asu": sc.array(
                dims=["row"],
                values=[
                    "[1, 2, 3]",
                    "[4, 5, 6]",
                    "[7, 8, 9]",
                    "[10, 11, 12]",
                    "[13, 14, 15]",
                    "[7, 8, 9]",
                    "[9, 8, 7]",
                ],
            ),
        },
    )
    da.variances = (
        sc.array(dims=["row"], values=[0.1, 0.2, 0.3, 0.4, 0.5, 0.31, 0.32]) ** 2
    )
    return da


def test_get_reference_bin_middle(nmx_data_array: sc.DataArray) -> None:
    """Test the middle bin."""

    binned = nmx_data_array.bin({"wavelength": 6})
    reference_wavelength = get_reference_wavelength(binned, reference_wavelength=None)

    ref_bin = get_reference_intensities(
        nmx_data_array.bin({"wavelength": 6}),
        reference_wavelength,
    )
    selected_idx = (2, 5, 6)
    assert all(
        ref_bin.data.values == [nmx_data_array.data.values[idx] for idx in selected_idx]
    )


@pytest.fixture
def reference_bin(nmx_data_array: sc.DataArray) -> ReferenceIntensities:
    binned = nmx_data_array.bin({"wavelength": 6})
    reference_wavelength = get_reference_wavelength(binned, reference_wavelength=None)

    return get_reference_intensities(
        binned,
        reference_wavelength,
    )


def test_reference_bin_scale_factor(reference_bin: ReferenceIntensities) -> None:
    """Test the scale factor for I."""
    scale_factor = estimate_scale_factor_per_hkl_asu_from_reference(reference_bin)
    expected_groups = [[7, 8, 9], [9, 8, 7]]

    assert len(scale_factor) == len(expected_groups)
    assert scale_factor.dim == "hkl_asu"
    for idx, group in enumerate(expected_groups):
        assert scale_factor.coords['hkl_asu'][idx].value == str(group)
